import numpy as np
import pandas as pd
from scipy.interpolate import interp1d

def calculate_PC(amplitude: list, sensitivity: float, gain: list, change_positions: list) -> np.ndarray:
    """
    Function that given input data returns the photocurrent spectra calculated
    with each 

    Parameters
    ----------
    amplitude : list of float
        input data of photocurrent
    gain : list of float
        vector with gains used
    sensitivity : float
        sensitivity of lockin
    change_positions : list of float
        vector with positions at which the gain has been changed

    Returns
    -------
    Amp: numpy array
        calculated amplitude photocurrent

    """
    
    Amp = []#vector of amplitudes
    sens = sensitivity #sensitivity 
    p = 0 # counter for initial position
    for i, j in enumerate(change_positions):
        amp = amplitude[p:j+1] #take that part of spectrum
        amp = amp*sens*1e-4 /gain[i] #adjust with gain and sens
        p = j+1
        Amp = np.concatenate((Amp, amp))
    return Amp

def lamp_spec(amplitude: list, wavelength: list, sensitivity: float, gain: list, change:list , diode_area: float) -> np.ndarray:
    """
    Function that takes into account a set of lamp data and returns the adjustment
    of the set with the sensitivity of photodiode

    Parameters
    ----------
    
    amplitude : list of float
        values of lamp amplitude   
    wavelength : list of float
        values of acquired wavelength
    sensitivity : float
        lockin sensitivity
    gain : list of float
        vector with gains used
    change : list of float
        vector with positions at which the gain has been changed
    
    Returns
    -------
    L1/A : numpy array
        vector with data adjusted with sensitivity

    """
    A = diode_area
    L1=[]
    
    #import photodiode sensitivities
    try:
        pdc = pd.read_csv("SensitivityPhotodiode.dat", sep="\t", names=["WL", "S"])
    except:
        print("You must have the file with diode sensitivity in the folder!")
        exit()
    #perform and interpolation to smooth conversion
    pdc_int = interp1d(pdc.WL, pdc.S)
    sens_new_WL = np.linspace(200,1000, 1000 )
    sens_new = pdc_int(sens_new_WL)
    #calculate spectrum of lamp
    PC = calculate_PC(amplitude, sensitivity, gain, change)
    #now multiply sensitivity
    for i in range(len(wavelength)):
        for j in range(len(sens_new_WL)):
            if (wavelength[i] >= sens_new_WL[j] and wavelength[i] < sens_new_WL[j+1]):
                L1.append(PC[i] /( 1e-3 * sens_new[j]))
                break
            else: continue
    L1 = np.asarray(L1)
    #return adjusted values in W/cm^2
    return L1/A


import numpy as np
import matplotlib.pyplot as plt
from scipy.integrate import trapz
from scipy.optimize import curve_fit
import pandas as pd
from datetime import datetime
import time
import AmorphSC.in_out as io

def solve_Runge_Kutta(x0: float, f, y0: float, n_steps:int , maxs: float, mins: float) -> list:
    """
    Parameters
    ----------
    x0 : float
        initial point for x vector
    f : function
        function on left hand side of equation: dy/dx = f(x,y)
    y0 : float
        initial point of y
    n_steps : int
        number of steps to perform
    maxs : float
        maximum value of x
    mins : float
        minimum value of x

    Returns
    -------
    y : array of float
        propagated solution with RK4 algorithm

    """
    #First define step size
    h = (maxs -mins )/n_steps
    
    #Define an empty vector to allocate values of the solution
    y = []
    #Initialize solution vector
    y.append(y0)
    
    #For n_steps times propagate the solution with 4th order RK method
    for i in range(1,n_steps):
        #Calculation of coefficients
        k1 = h * f(x0, y[i-1])
        k2 = h * f(x0+h/2 , y[i-1] + k1/2)
        k3 = h * f(x0+h/2 , y[i-1] + k2/2)
        k4 = h * f(x0+h , y[i-1]+k3)
        yn1 = y[i-1]+k1/6 +k2/3 + k3/3 +k4/6
        y.append(yn1)
        x0 = x0+h
        
    return y

def solve_Poisson_Tip(phit_min: float, phit_max: float ,ND: float,
                      Et: float, VFB: float, W: float, L: float, t: float, Cox: float, 
                      N_points: int) -> pd.core.frame.DataFrame:
    """
    Parameters
    ----------
    phit_min : float
        minimum value of phit
    phit_max : float
        maximum value of phit
    ND : float
        doping density
    Et : float
        band tail width
    VFB : float
        flat band potential
    W : float
        width of channel
    L : float
        length of channel
    t : float
        thickness of semiconducing layer
    Cox : float
        oxide capacitance of the transistor
    N_points : int
        number of points for RK algorithm

    Returns
    -------
    Return a dataframe with all the calculated quantities inside

    """
   
    

    #First derivative function
    def dzdx(y,y0):
        if y0>0:
            return k*(np.exp(O*y)-1)
        if y0<0:
            return k*(1-np.exp(O*y))

    def rho(y, y0):
       return q0*ND*(np.exp(O*y)-1)
   

    eps0=8.85e-12 #F/m
    eps=10
    q0=1.6e-19 #C

    #Calculation of parameters
    k = q0*ND /(eps*eps0)
    O = 1/Et

    N = N_points
    phit = np.linspace(phit_min,phit_max,N)
    mins=0
    X = np.linspace(0,t,N)

    Phi0 = []
    Rho = []
    Phi = []

    for j in phit:
        out = solve_Runge_Kutta(mins, dzdx, j, N, t, mins)
        out = np.asarray(out)
        Phi.append(out)
        Phi0.append(out[N-1])
        rho_val = rho(out, j)
        rho_val = trapz(rho_val,X)
        Rho.append(rho_val)
        
    Rho = np.asarray(Rho)
    Phi0 = np.asarray(Phi0)

    VG = VFB + Phi0 + Rho/Cox
    
    c = np.gradient(Rho) / np.gradient(VG)
    
    
    lib = {'Phit': phit, 'VG':VG, 'Rho':Rho, 'RhoAbs': abs(Rho),'Phi0':Phi0, 'c':c }
    out = pd.DataFrame(lib)
    
    return out

def fit_Poisson_Tip(phit ,ND, Et, VFB, t, Cox):
    """
    Function used to numerically fit a system and solve Poisson's equation

    Parameters
    ----------
    phit : list
        list with values of phit.
    ND : float
        doping density (to be fitted).
    Et : float
        band taild width (tfb).
    VFB : float
        flat band potential (tbf).
    t : float
        thickness of semiconducting layer (to provide).
    Cox : float
        oxide layer capacitance per unit area (tp).

    Returns
    -------
    VG : numpy.ndarray
        Values of gate voltage calculated solving poisson equation and adapted
        with fitting procedure.

    """

    #First derivative function
    def dzdx(y,y0):
        if y0>0:
            return k*(np.exp(O*y)-1)
        if y0<0:
            return k*(1-np.exp(O*y))

    def rho(y, y0):
       return q0*ND*(np.exp(O*y)-1)

        

    """PARAMETERS"""

    eps0=8.85e-12 #F/m
    eps=10
    q0=1.6e-19 #C
     
    phit = np.asarray(phit)

    #Calculation of parameters
    k = q0*ND /(eps*eps0)
    O = 1/Et

    N = 200
    mins=0
    X = np.linspace(0,t,N)
  

    Phi0 = []
    Rho = []
    Phi = []

    for j in phit:
        out = solve_Runge_Kutta(mins, dzdx, j, N, t, mins)
        out = np.asarray(out)
        Phi.append(out)
        Phi0.append(out[N-1])
        rho_val = rho(out, j)
        rho_val = trapz(rho_val,X)
        Rho.append(rho_val)
        
    Rho = np.asarray(Rho)
    Phi0 = np.asarray(Phi0)

    VG = VFB + Phi0 + Rho/Cox
    
    print(datetime.now())
    
    return VG

def fit_KPFM_curve(VG: list, phit: list, t: float, Cox: float, p0=[1,1,1]) -> np.ndarray:
    """
    

    Parameters
    ----------
    VG : list
        Values of measured gate voltage.
    phit : list
        Values of measured KPFM potential.
    t : float
        Thickness of semiconducting layer.
    Cox : float
        Oxide capacitance of semiconducting layer.
    p0 : list, OPTIONAL
        Array of initial parameters for ND, Et and VFB

    Returns
    -------
    Parameters fitted

    """
    
    start_time = time.time() #start time counting
    
    t_set = t
    Cox_set = Cox
    
    reduced_f = lambda phit, ND, Et, VFB: fit_Poisson_Tip(phit, ND, Et, VFB, t_set, Cox_set)
    
    #Fit data setting to completely bounded parameters t and Cox since they
    #are provided from the user
    popt, pcov = curve_fit(reduced_f, xdata = phit, ydata = VG,
                           p0 = p0)
                           
    
    err = np.sqrt(np.diag(pcov)) #calculate errors on parameters

    #Assing parameters
    ND = popt[0]
    Et = popt[1]
    VFB = popt[2]

    #Stop time acquisition
    end_time = time.time()

    total = (end_time - start_time) #in minutes
    
    #Print informations on console
    print("ND = ", io.s4(ND), " +- ", io.s4(err[0]))
    print("Et = ", io.s4(Et), " +- ", io.s4(err[1]))
    print("VFB = ", io.s4(VFB), " +- ", io.s4(err[2]))
    print("Time used (mins): ", total)
    
    
    #Output File
    names = ["ND", "Et", "VFB", "Time"]
    values = [io.s4(ND), io.s4(Et), io.s4(VFB), total ]
    errors = [io.s4(err[0]), io.s4(err[1]), io.s4(err[2]),0]

    lib_out = {'Name':names, 'Value':values, 'Error':errors}
    data_out = pd.DataFrame(lib_out)
    data_out.to_csv("Data_Out_Fit.txt", sep="\t")

    return ND, Et, VFB
    


import pandas as pd
from scipy.integrate import cumtrapz
import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import curve_fit

def calculate_capacitance(current: list, frequency: float, V_out: float, W: float, L: float, n_tft=500) -> list:
    """
    Function that calculates capacitance from I-V data recorded. Since data are
    acquired as AC signal then C = I/(2*pi*f*V_out)

    Parameters
    ----------
    current : list of float
        Current value registered from lock-in amplifier
    frequency : float
        DFrequency of acquisition
    V_out : float
        RMS of peak-to-peak value of base voltage in lock-in amplifier
    W : float
        Width of the transistor's channel in meters
    L : float
        Length of the transistor's channel in meters
    n_tft : int, optional
        Numer of trasistors in the sample. The default is 500.

    Returns
    -------
    C : float
        Values of capacitance

    """
    C = current /(2*3.1415*frequency*V_out)
    C = C-min(C) #remove parasite capacitance
    return C

def calculate_DOS(capacitance: list, oxide_capacitance: float, W: float, L: float, t: float) -> list:
    """
    Function that calculated density of states from capacitance data

    Parameters
    ----------
    capacitance : list of float
        Capacitance data
    oxide_capacitance : float
        Specific capacitance of oxide dielectric
    W : float
        Width of transistor channel in meters
    L : float
        Length of transistor channel in meteres
    t : float
        Thickness of dielectric

    Returns
    -------
    list of float
        Values of density of states

    """
    q0 = 1.6e-19 #electron charge
    CD = (capacitance*oxide_capacitance) / (oxide_capacitance-capacitance)
    return (CD /(q0*W*L*t))

def calculate_energy_range(capacitance: list, oxide_capacitance: float, voltage: list, correction=0, init = 0) -> list:
    """
    Function that calculates energy (or surface potential) from data

    Parameters
    ----------
    capacitance : list of float
        Values of measured capacitance
    oxide_capacitance : float
        Specific capacitance of oxide layer
    voltage : list of float
        Values of voltage offset used in CV measure
    correction : float, optional
        Optional correction to energy scale (account for flat band
                                             voltage). The default is 0.
    init : float, optional
        Initial value for integration. The default is 0.

    Returns
    -------
    E : list of float
        Values of energy range.

    """
    E = cumtrapz(1-capacitance/oxide_capacitance,voltage, initial = init)+correction
    return E

def find_fit_interval(E: list, g: list , FitLeft: float, FitRight: float ):
    """
    Function that finds intervals for fitting of DOS vs Energy

    Parameters
    ----------
    E : list of float
        Energy values
    g : list of float
        DOS values
    FitLeft : float
        Left margin to fit data
    FitRight : float
        DRight margin to fit data

    Returns
    -------
    E : list of float
        Energy range between Left and Right
    g : list of float
        DOS range between Left and Right

    """
    fit_interval = np.where((E>=FitLeft) & (E<=FitRight))
    fit_interval = fit_interval[0]
    g = g[fit_interval]
    E = E[fit_interval]
    return E, g

def c_thick_func(phi, k, omega, delta):
    """
    Analytical function used to fit CV data in function of surface potential.
    Paramters are k, omega and delta while phi is the variable.
    """
    epsilon_0 = 8.85e-12
    epsilon=10
    output = []
    for x in range(len(phi)):
        num = np.exp(omega*(phi[x]-delta)) -1
        den = np.sqrt(2*(np.exp(omega*(phi[x]-delta))-omega*(phi[x]-delta)-1))
        if phi[x]-delta >=  0:
            p1 = epsilon*epsilon_0 * np.sqrt(k/omega)
        elif phi[x]-delta < 0 :
            p1 = -epsilon*epsilon_0 * np.sqrt(k/omega)
        output.append(p1*num/den)   

    return output



import glob
import pandas as pd

def import_file(prefix_file: str, names: list, separator="\t", header=0) -> list:
    """
    Function that given a prefix file composed by path and prexif can import
    all files with that prefix in the given folder.
    

    Parameters
    ----------
    prefix_file : string
        common prexif of files that have to be imported
    names : list of string
        vector of names of file columns
    sep : string, optional
        Separator of each column. The default is "\t".
    header : int, optional
        Number of line of the header. The default is 0.

    Returns
    -------
    char : 
        List with imported files.

    """
    
    files_char = glob.glob(prefix_file)
    char = []
    for j in files_char:
        csv = pd.read_csv(j, sep = separator, header = header, names=names)
        char.append(csv)
        print(j)
    return char


def s4(a):
    return '{:.4e}'.format(a)

def s0(a):
    return '{:.0e}'.format(a)

def s1(a):
    return '{:.1e}'.format(a)



from distutils.core import setup

setup(
  name = 'AmorphSC',         # How you named your package folder (MyLib)
  packages = ['AmorphSC'],   # Chose the same as "name"
  version = '1.0',      # Start with a small number and increase it with every change you make
  license='GNU',        # Chose a license from here: https://help.github.com/articles/licensing-a-repository
  description = 'Package to permorm data analysis on amorphous semiconductors',   # Give a short description about your library
  author = 'Luca Fabbri',                   # Type in your name
  author_email = 'luca.fabbri98@outlook.it',      # Type in your E-Mail
  url = 'https://github.com/lfabbri98/AmorphSC',   # Provide either the link to your github or to your website
  download_url = 'https://github.com/user/reponame/archive/v_01.tar.gz',    # I explain this later on
  keywords = ['semiconductors', 'amorphous', 'KPFM', 'AFM','photocurrent'],   # Keywords that define your package best
  install_requires=[            # I get to this in a second
          'pandas',
          'matplotlib',
          'scipy',
          'numpy',
      ],
)
